from __future__ import annotations

import pytest

from PIL import Image, ImageDraw, ImageFont

from .helper import (
    assert_image_equal_tofile,
    assert_image_similar_tofile,
    has_feature_version,
    skip_unless_feature,
)

FONT_SIZE = 20
FONT_PATH = "Tests/fonts/DejaVuSans/DejaVuSans.ttf"

pytestmark = skip_unless_feature("raqm")


def test_english() -> None:
    # smoke test, this should not fail
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)
    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "TEST", font=ttf, fill=500, direction="ltr")


def test_complex_text() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "اهلا عمان", font=ttf, fill=500)

    target = "Tests/images/test_text.png"
    assert_image_similar_tofile(im, target, 0.5)


def test_y_offset() -> None:
    ttf = ImageFont.truetype("Tests/fonts/NotoNastaliqUrdu-Regular.ttf", FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "العالم العربي", font=ttf, fill=500)

    target = "Tests/images/test_y_offset.png"
    assert_image_similar_tofile(im, target, 1.7)


def test_complex_unicode_text() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "السلام عليكم", font=ttf, fill=500)

    target = "Tests/images/test_complex_unicode_text.png"
    assert_image_similar_tofile(im, target, 0.5)

    ttf = ImageFont.truetype("Tests/fonts/KhmerOSBattambang-Regular.ttf", FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "លោកុប្បត្តិ", font=ttf, fill=500)

    target = "Tests/images/test_complex_unicode_text2.png"
    assert_image_similar_tofile(im, target, 2.33)


def test_text_direction_rtl() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "English عربي", font=ttf, fill=500, direction="rtl")

    target = "Tests/images/test_direction_rtl.png"
    assert_image_similar_tofile(im, target, 0.5)


def test_text_direction_ltr() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "سلطنة عمان Oman", font=ttf, fill=500, direction="ltr")

    target = "Tests/images/test_direction_ltr.png"
    assert_image_similar_tofile(im, target, 0.5)


def test_text_direction_rtl2() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "Oman سلطنة عمان", font=ttf, fill=500, direction="rtl")

    target = "Tests/images/test_direction_ltr.png"
    assert_image_similar_tofile(im, target, 0.5)


def test_text_direction_ttb() -> None:
    ttf = ImageFont.truetype("Tests/fonts/NotoSansJP-Regular.otf", FONT_SIZE)

    im = Image.new(mode="RGB", size=(100, 300))
    draw = ImageDraw.Draw(im)
    if not has_feature_version("raqm", "0.7"):
        pytest.skip("libraqm 0.7 or greater not available")
    draw.text((0, 0), "English あい", font=ttf, fill=500, direction="ttb")

    target = "Tests/images/test_direction_ttb.png"
    assert_image_similar_tofile(im, target, 2.8)


def test_text_direction_ttb_stroke() -> None:
    ttf = ImageFont.truetype("Tests/fonts/NotoSansJP-Regular.otf", 50)

    im = Image.new(mode="RGB", size=(100, 300))
    draw = ImageDraw.Draw(im)
    if not has_feature_version("raqm", "0.7"):
        pytest.skip("libraqm 0.7 or greater not available")
    draw.text(
        (27, 27),
        "あい",
        font=ttf,
        fill=500,
        direction="ttb",
        stroke_width=2,
        stroke_fill="#0f0",
    )

    target = "Tests/images/test_direction_ttb_stroke.png"
    assert_image_similar_tofile(im, target, 19.4)


def test_ligature_features() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "filling", font=ttf, fill=500, features=["-liga"])
    target = "Tests/images/test_ligature_features.png"
    assert_image_similar_tofile(im, target, 0.5)

    liga_bbox = ttf.getbbox("fi", features=["-liga"])
    assert liga_bbox == (0, 4, 13, 19)


def test_kerning_features() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "TeToAV", font=ttf, fill=500, features=["-kern"])

    target = "Tests/images/test_kerning_features.png"
    assert_image_similar_tofile(im, target, 0.5)


def test_arabictext_features() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text(
        (0, 0),
        "اللغة العربية",
        font=ttf,
        fill=500,
        features=["-fina", "-init", "-medi"],
    )

    target = "Tests/images/test_arabictext_features.png"
    assert_image_similar_tofile(im, target, 0.5)


def test_x_max_and_y_offset() -> None:
    ttf = ImageFont.truetype("Tests/fonts/ArefRuqaa-Regular.ttf", 40)

    im = Image.new(mode="RGB", size=(50, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "لح", font=ttf, fill=500)

    target = "Tests/images/test_x_max_and_y_offset.png"
    assert_image_similar_tofile(im, target, 3.8)


def test_language() -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)

    im = Image.new(mode="RGB", size=(300, 100))
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), "абвг", font=ttf, fill=500, language="sr")

    target = "Tests/images/test_language.png"
    assert_image_similar_tofile(im, target, 0.5)


@pytest.mark.parametrize("mode", ("L", "1"))
@pytest.mark.parametrize(
    "text, direction, expected",
    (
        ("سلطنة عمان Oman", None, 173.703125),
        ("سلطنة عمان Oman", "ltr", 173.703125),
        ("Oman سلطنة عمان", "rtl", 173.703125),
        ("English عربي", "rtl", 123.796875),
        ("test", "ttb", 80.0),
    ),
    ids=("None", "ltr", "rtl2", "rtl", "ttb"),
)
def test_getlength(
    mode: str, text: str, direction: str | None, expected: float
) -> None:
    ttf = ImageFont.truetype(FONT_PATH, FONT_SIZE)
    im = Image.new(mode, (1, 1), 0)
    d = ImageDraw.Draw(im)

    if direction == "ttb" and not has_feature_version("raqm", "0.7"):
        pytest.skip("libraqm 0.7 or greater not available")
    assert d.textlength(text, ttf, direction) == expected


@pytest.mark.parametrize("mode", ("L", "1"))
@pytest.mark.parametrize("direction", ("ltr", "ttb"))
@pytest.mark.parametrize(
    "text",
    ("i" + ("\u030c" * 15) + "i", "i" + "\u032c" * 15 + "i", "\u035cii", "i\u0305i"),
    ids=("caron-above", "caron-below", "double-breve", "overline"),
)
def test_getlength_combine(mode: str, direction: str, text: str) -> None:
    if text == "i\u0305i" and direction == "ttb":
        pytest.skip("fails with this font")

    ttf = ImageFont.truetype("Tests/fonts/NotoSans-Regular.ttf", 48)

    if direction == "ttb" and not has_feature_version("raqm", "0.7"):
        pytest.skip("libraqm 0.7 or greater not available")
    target = ttf.getlength("ii", mode, direction)
    actual = ttf.getlength(text, mode, direction)

    assert actual == target


@pytest.mark.parametrize("anchor", ("lt", "mm", "rb", "sm"))
def test_anchor_ttb(anchor: str) -> None:
    text = "f"
    path = f"Tests/images/test_anchor_ttb_{text}_{anchor}.png"
    f = ImageFont.truetype("Tests/fonts/NotoSans-Regular.ttf", 120)

    im = Image.new("RGB", (200, 400), "white")
    d = ImageDraw.Draw(im)
    d.line(((0, 200), (200, 200)), "gray")
    d.line(((100, 0), (100, 400)), "gray")
    if not has_feature_version("raqm", "0.7"):
        pytest.skip("libraqm 0.7 or greater not available")
    d.text((100, 200), text, fill="black", anchor=anchor, direction="ttb", font=f)

    assert_image_similar_tofile(im, path, 1)  # fails at 5


combine_tests = (
    # extends above (e.g. issue #4553)
    ("caron", "a\u030c\u030c\u030c\u030c\u030cb", None, None, 0.08),
    ("caron_la", "a\u030c\u030c\u030c\u030c\u030cb", "la", None, 0.08),
    ("caron_lt", "a\u030c\u030c\u030c\u030c\u030cb", "lt", None, 0.08),
    ("caron_ls", "a\u030c\u030c\u030c\u030c\u030cb", "ls", None, 0.08),
    ("caron_ttb", "ca" + ("\u030c" * 15) + "b", None, "ttb", 0.3),
    ("caron_ttb_lt", "ca" + ("\u030c" * 15) + "b", "lt", "ttb", 0.3),
    # extends below
    ("caron_below", "a\u032c\u032c\u032c\u032c\u032cb", None, None, 0.02),
    ("caron_below_ld", "a\u032c\u032c\u032c\u032c\u032cb", "ld", None, 0.02),
    ("caron_below_lb", "a\u032c\u032c\u032c\u032c\u032cb", "lb", None, 0.02),
    ("caron_below_ls", "a\u032c\u032c\u032c\u032c\u032cb", "ls", None, 0.02),
    ("caron_below_ttb", "a" + ("\u032c" * 15) + "b", None, "ttb", 0.03),
    ("caron_below_ttb_lb", "a" + ("\u032c" * 15) + "b", "lb", "ttb", 0.03),
    # extends to the right (e.g. issue #3745)
    ("double_breve_below", "a\u035ci", None, None, 0.02),
    ("double_breve_below_ma", "a\u035ci", "ma", None, 0.02),
    ("double_breve_below_ra", "a\u035ci", "ra", None, 0.02),
    ("double_breve_below_ttb", "a\u035cb", None, "ttb", 0.02),
    ("double_breve_below_ttb_rt", "a\u035cb", "rt", "ttb", 0.02),
    ("double_breve_below_ttb_mt", "a\u035cb", "mt", "ttb", 0.02),
    ("double_breve_below_ttb_st", "a\u035cb", "st", "ttb", 0.02),
    # extends to the left (fail=0.064)
    ("overline", "i\u0305", None, None, 0.02),
    ("overline_la", "i\u0305", "la", None, 0.02),
    ("overline_ra", "i\u0305", "ra", None, 0.02),
    ("overline_ttb", "i\u0305", None, "ttb", 0.02),
    ("overline_ttb_rt", "i\u0305", "rt", "ttb", 0.02),
    ("overline_ttb_mt", "i\u0305", "mt", "ttb", 0.02),
    ("overline_ttb_st", "i\u0305", "st", "ttb", 0.02),
)


# this tests various combining characters for anchor alignment and clipping
@pytest.mark.parametrize(
    "name, text, anchor, direction, epsilon",
    combine_tests,
    ids=[r[0] for r in combine_tests],
)
def test_combine(
    name: str, text: str, direction: str | None, anchor: str | None, epsilon: float
) -> None:
    path = f"Tests/images/test_combine_{name}.png"
    f = ImageFont.truetype("Tests/fonts/NotoSans-Regular.ttf", 48)

    im = Image.new("RGB", (400, 400), "white")
    d = ImageDraw.Draw(im)
    d.line(((0, 200), (400, 200)), "gray")
    d.line(((200, 0), (200, 400)), "gray")
    if direction == "ttb" and not has_feature_version("raqm", "0.7"):
        pytest.skip("libraqm 0.7 or greater not available")
    d.text((200, 200), text, fill="black", anchor=anchor, direction=direction, font=f)

    assert_image_similar_tofile(im, path, epsilon)


@pytest.mark.parametrize(
    "anchor, align",
    (
        ("lm", "left"),  # pass with getsize
        ("lm", "center"),  # fail at 2.12
        ("lm", "right"),  # fail at 2.57
        ("mm", "left"),  # fail at 2.12
        ("mm", "center"),  # pass with getsize
        ("mm", "right"),  # fail at 2.12
        ("rm", "left"),  # fail at 2.57
        ("rm", "center"),  # fail at 2.12
        ("rm", "right"),  # pass with getsize
    ),
)
def test_combine_multiline(anchor: str, align: str) -> None:
    # test that multiline text uses getlength, not getsize or getbbox

    path = f"Tests/images/test_combine_multiline_{anchor}_{align}.png"
    f = ImageFont.truetype("Tests/fonts/NotoSans-Regular.ttf", 48)
    text = "i\u0305\u035c\ntext"  # i with overline and double breve, and a word

    im = Image.new("RGB", (400, 400), "white")
    d = ImageDraw.Draw(im)
    d.line(((0, 200), (400, 200)), "gray")
    d.line(((200, 0), (200, 400)), "gray")
    bbox = d.multiline_textbbox((200, 200), text, anchor=anchor, font=f, align=align)
    d.rectangle(bbox, outline="red")
    d.multiline_text((200, 200), text, "black", anchor=anchor, font=f, align=align)

    assert_image_similar_tofile(im, path, 0.015)


def test_combine_multiline_ttb() -> None:
    path = "Tests/images/test_combine_multiline_ttb.png"
    f = ImageFont.truetype("Tests/fonts/NotoSans-Regular.ttf", 48)
    text = "te\nxt"

    im = Image.new("RGB", (400, 400), "white")
    d = ImageDraw.Draw(im)
    d.line(((0, 200), (400, 200)), "gray")
    d.line(((200, 0), (200, 400)), "gray")
    bbox = d.multiline_textbbox((200, 200), text, f, direction="ttb")
    d.rectangle(bbox, outline="red")
    d.multiline_text((200, 200), text, "black", f, direction="ttb")

    assert_image_equal_tofile(im, path)


def test_anchor_invalid_ttb() -> None:
    font = ImageFont.truetype(FONT_PATH, FONT_SIZE)
    im = Image.new("RGB", (100, 100), "white")
    d = ImageDraw.Draw(im)
    d.font = font

    for anchor in ["", "l", "a", "lax", "xa", "la", "ls", "ld", "lx"]:
        with pytest.raises(ValueError):
            font.getmask2("hello", anchor=anchor, direction="ttb")
        with pytest.raises(ValueError):
            font.getbbox("hello", anchor=anchor, direction="ttb")
        with pytest.raises(ValueError):
            d.text((0, 0), "hello", anchor=anchor, direction="ttb")
        with pytest.raises(ValueError):
            d.textbbox((0, 0), "hello", anchor=anchor, direction="ttb")
        with pytest.raises(ValueError):
            d.multiline_text((0, 0), "foo\nbar", anchor=anchor, direction="ttb")
        with pytest.raises(ValueError):
            d.multiline_textbbox((0, 0), "foo\nbar", anchor=anchor, direction="ttb")



from __future__ import annotations

import warnings

import pytest

from PIL import Image

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=DeprecationWarning)
    from PIL import ImageQt

from .helper import assert_image_equal, hopper

pytestmark = pytest.mark.skipif(
    not ImageQt.qt_is_installed, reason="Qt bindings are not installed"
)

ims: list[Image.Image] = []


def setup_module() -> None:
    ims.append(hopper())
    ims.append(Image.open("Tests/images/transparent.png"))
    ims.append(Image.open("Tests/images/7x13.png"))


def teardown_module() -> None:
    for im in ims:
        im.close()


def roundtrip(expected: Image.Image) -> None:
    # PIL -> Qt
    intermediate = expected.toqimage()
    # Qt -> PIL
    result = ImageQt.fromqimage(intermediate)

    if intermediate.hasAlphaChannel():
        assert_image_equal(result, expected.convert("RGBA"))
    else:
        assert_image_equal(result, expected.convert("RGB"))


def test_sanity_1() -> None:
    for im in ims:
        roundtrip(im.convert("1"))


def test_sanity_rgb() -> None:
    for im in ims:
        roundtrip(im.convert("RGB"))


def test_sanity_rgba() -> None:
    for im in ims:
        roundtrip(im.convert("RGBA"))


def test_sanity_l() -> None:
    for im in ims:
        roundtrip(im.convert("L"))


def test_sanity_p() -> None:
    for im in ims:
        roundtrip(im.convert("P"))



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import GdImageFile, Image, UnidentifiedImageError, _binary

from .helper import assert_image_similar_tofile

TEST_GD_FILE = "Tests/images/hopper.gd"


def test_sanity() -> None:
    with GdImageFile.open(TEST_GD_FILE) as im:
        assert im.size == (128, 128)
        assert im.format == "GD"
        assert_image_similar_tofile(im.convert("RGB"), "Tests/images/hopper.jpg", 14)


def test_transparency() -> None:
    with open(TEST_GD_FILE, "rb") as fp:
        data = bytearray(fp.read())
    data[7:11] = b"\x00\x00\x00\x05"
    with GdImageFile.open(BytesIO(data)) as im:
        assert im.info["transparency"] == 5


def test_bad_mode() -> None:
    with pytest.raises(ValueError):
        GdImageFile.open(TEST_GD_FILE, "bad mode")


def test_decompression_bomb() -> None:
    b = BytesIO(_binary.o16be(65535) + _binary.o16be(50000) + _binary.o16be(50000))
    with pytest.raises(Image.DecompressionBombError):
        GdImageFile.open(b)


def test_invalid_file() -> None:
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(UnidentifiedImageError):
        GdImageFile.open(invalid_file)



# Reproductions/tests for OOB read errors in FliDecode.c

# When run in python, all of these images should fail for
# one reason or another, either as a buffer overrun,
# unrecognized datastream, or truncated image file.
# There shouldn't be any segfaults.
#
# if run like
# `valgrind --tool=memcheck python check_jp2_overflow.py  2>&1 | grep Decode.c`
# the output should be empty. There may be python issues
# in the valgrind especially if run in a debug python
# version.
from __future__ import annotations

from PIL import Image

repro = ("00r0_gray_l.jp2", "00r1_graya_la.jp2")

for path in repro:
    with Image.open(path) as im:
        try:
            im.load()
        except Exception as msg:
            print(msg)



from __future__ import annotations

import io
from pathlib import Path

import pytest

from PIL import Image, ImagePalette, PaletteFile

from .helper import assert_image_equal, assert_image_equal_tofile


def test_sanity() -> None:
    palette = ImagePalette.ImagePalette("RGB", list(range(256)) * 3)
    assert len(palette.colors) == 256


def test_reload() -> None:
    with Image.open("Tests/images/hopper.gif") as im:
        original = im.copy()
        assert im.palette is not None
        im.palette.dirty = 1
        assert_image_equal(im.convert("RGB"), original.convert("RGB"))


def test_save_fp() -> None:
    palette = ImagePalette.ImagePalette()
    with io.StringIO() as fp:
        palette.save(fp)
        assert not fp.closed


def test_getcolor() -> None:
    palette = ImagePalette.ImagePalette()
    assert len(palette.palette) == 0
    assert len(palette.colors) == 0

    test_map = {}
    for i in range(256):
        test_map[palette.getcolor((i, i, i))] = i
    assert len(test_map) == 256

    # Colors can be converted between RGB and RGBA
    rgba_palette = ImagePalette.ImagePalette("RGBA")
    assert rgba_palette.getcolor((0, 0, 0)) == rgba_palette.getcolor((0, 0, 0, 255))

    assert palette.getcolor((0, 0, 0)) == palette.getcolor((0, 0, 0, 255))

    # An error is raised when the palette is full
    with pytest.raises(ValueError):
        palette.getcolor((1, 2, 3))
    # But not if the image is not using one of the palette entries
    palette.getcolor((1, 2, 3), image=Image.new("P", (1, 1)))

    # Test unknown color specifier
    with pytest.raises(ValueError):
        palette.getcolor("unknown")  # type: ignore[arg-type]


def test_getcolor_rgba() -> None:
    palette = ImagePalette.ImagePalette("RGBA", (1, 2, 3, 4))
    palette.getcolor((5, 6, 7, 8))
    assert palette.palette == b"\x01\x02\x03\x04\x05\x06\x07\x08"


def test_getcolor_rgba_color_rgb_palette() -> None:
    palette = ImagePalette.ImagePalette("RGB")

    # Opaque RGBA colors are converted
    assert palette.getcolor((0, 0, 0, 255)) == palette.getcolor((0, 0, 0))

    with pytest.raises(ValueError):
        palette.getcolor((0, 0, 0, 128))


@pytest.mark.parametrize(
    "index, palette",
    [
        # Test when the palette is not full
        (0, ImagePalette.ImagePalette()),
        # Test when the palette is full
        (255, ImagePalette.ImagePalette("RGB", list(range(256)) * 3)),
    ],
)
def test_getcolor_not_special(index: int, palette: ImagePalette.ImagePalette) -> None:
    im = Image.new("P", (1, 1))

    # Do not use transparency index as a new color
    im.info["transparency"] = index
    index1 = palette.getcolor((0, 0, 0), im)
    assert index1 != index

    # Do not use background index as a new color
    im.info["background"] = index1
    index2 = palette.getcolor((0, 0, 1), im)
    assert index2 not in (index, index1)


def test_file(tmp_path: Path) -> None:
    palette = ImagePalette.ImagePalette("RGB", list(range(256)) * 3)

    f = str(tmp_path / "temp.lut")

    palette.save(f)

    lut = ImagePalette.load(f)

    # load returns raw palette information
    assert len(lut[0]) == 768
    assert lut[1] == "RGB"

    p = ImagePalette.raw(lut[1], lut[0])
    assert isinstance(p, ImagePalette.ImagePalette)
    assert p.palette == palette.tobytes()


def test_make_linear_lut() -> None:
    # Arrange
    black = 0
    white = 255

    # Act
    lut = ImagePalette.make_linear_lut(black, white)

    # Assert
    assert isinstance(lut, list)
    assert len(lut) == 256
    # Check values
    for i in range(len(lut)):
        assert lut[i] == i


def test_make_linear_lut_not_yet_implemented() -> None:
    # Update after FIXME
    # Arrange
    black = 1
    white = 255

    # Act
    with pytest.raises(NotImplementedError):
        ImagePalette.make_linear_lut(black, white)


def test_make_gamma_lut() -> None:
    # Arrange
    exp = 5

    # Act
    lut = ImagePalette.make_gamma_lut(exp)

    # Assert
    assert isinstance(lut, list)
    assert len(lut) == 256
    # Check a few values
    assert lut[0] == 0
    assert lut[63] == 0
    assert lut[127] == 8
    assert lut[191] == 60
    assert lut[255] == 255


def test_rawmode_valueerrors(tmp_path: Path) -> None:
    # Arrange
    palette = ImagePalette.raw("RGB", list(range(256)) * 3)

    # Act / Assert
    with pytest.raises(ValueError):
        palette.tobytes()
    with pytest.raises(ValueError):
        palette.getcolor((1, 2, 3))
    f = str(tmp_path / "temp.lut")
    with pytest.raises(ValueError):
        palette.save(f)


def test_getdata() -> None:
    # Arrange
    data_in = list(range(256)) * 3
    palette = ImagePalette.ImagePalette("RGB", data_in)

    # Act
    mode, data_out = palette.getdata()

    # Assert
    assert mode == "RGB"


def test_rawmode_getdata() -> None:
    # Arrange
    data_in = list(range(256)) * 3
    palette = ImagePalette.raw("RGB", data_in)

    # Act
    rawmode, data_out = palette.getdata()

    # Assert
    assert rawmode == "RGB"
    assert data_in == data_out


def test_2bit_palette(tmp_path: Path) -> None:
    # issue #2258, 2 bit palettes are corrupted.
    outfile = str(tmp_path / "temp.png")

    rgb = b"\x00" * 2 + b"\x01" * 2 + b"\x02" * 2
    img = Image.frombytes("P", (6, 1), rgb)
    img.putpalette(b"\xff\x00\x00\x00\xff\x00\x00\x00\xff")  # RGB
    img.save(outfile, format="PNG")

    assert_image_equal_tofile(img, outfile)


def test_getpalette() -> None:
    b = io.BytesIO(b"0 1\n1 2 3 4")
    p = PaletteFile.PaletteFile(b)

    palette, rawmode = p.getpalette()
    assert palette[:6] == b"\x01\x01\x01\x02\x03\x04"
    assert rawmode == "RGB"


def test_invalid_palette() -> None:
    with pytest.raises(OSError):
        ImagePalette.load("Tests/images/hopper.jpg")

    b = io.BytesIO(b"1" * 101)
    with pytest.raises(SyntaxError, match="bad palette file"):
        PaletteFile.PaletteFile(b)



from __future__ import annotations

import zlib
from io import BytesIO

import pytest

from PIL import Image, ImageFile, PngImagePlugin

TEST_FILE = "Tests/images/png_decompression_dos.png"


def test_ignore_dos_text(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ImageFile, "LOAD_TRUNCATED_IMAGES", True)

    with Image.open(TEST_FILE) as im:
        im.load()

        assert isinstance(im, PngImagePlugin.PngImageFile)
        for s in im.text.values():
            assert len(s) < 1024 * 1024, "Text chunk larger than 1M"

        for s in im.info.values():
            assert len(s) < 1024 * 1024, "Text chunk larger than 1M"


def test_dos_text() -> None:
    try:
        im = Image.open(TEST_FILE)
        im.load()
    except ValueError as msg:
        assert msg, "Decompressed Data Too Large"
        return

    assert isinstance(im, PngImagePlugin.PngImageFile)
    for s in im.text.values():
        assert len(s) < 1024 * 1024, "Text chunk larger than 1M"


def test_dos_total_memory() -> None:
    im = Image.new("L", (1, 1))
    compressed_data = zlib.compress(b"a" * 1024 * 1023)

    info = PngImagePlugin.PngInfo()

    for x in range(64):
        info.add_text(f"t{x}", compressed_data, zip=True)
        info.add_itxt(f"i{x}", compressed_data, zip=True)

    b = BytesIO()
    im.save(b, "PNG", pnginfo=info)
    b.seek(0)

    try:
        im2 = Image.open(b)
    except ValueError as msg:
        assert "Too much memory" in str(msg)
        return

    total_len = 0
    assert isinstance(im2, PngImagePlugin.PngImageFile)
    for txt in im2.text.values():
        total_len += len(txt)
    assert total_len < 64 * 1024 * 1024, "Total text chunks greater than 64M"



from __future__ import annotations

import platform
import sys

from PIL import features


def test_wheel_modules() -> None:
    expected_modules = {"pil", "tkinter", "freetype2", "littlecms2", "webp", "avif"}

    if sys.platform == "win32":
        # tkinter is not available in cibuildwheel installed CPython on Windows
        try:
            import tkinter

            assert tkinter
        except ImportError:
            expected_modules.remove("tkinter")

        # libavif is not available on Windows for ARM64 architectures
        if platform.machine() == "ARM64":
            expected_modules.remove("avif")

    elif sys.platform == "ios":
        # tkinter is not available on iOS
        expected_modules.remove("tkinter")

    assert set(features.get_supported_modules()) == expected_modules


def test_wheel_codecs() -> None:
    expected_codecs = {"jpg", "jpg_2000", "zlib", "libtiff"}

    assert set(features.get_supported_codecs()) == expected_codecs


def test_wheel_features() -> None:
    expected_features = {
        "raqm",
        "fribidi",
        "harfbuzz",
        "libjpeg_turbo",
        "zlib_ng",
        "xcb",
    }

    if sys.platform == "win32":
        expected_features.remove("xcb")
    elif sys.platform == "ios":
        # Can't distribute raqm due to licensing, and there's no system version;
        # fribidi and harfbuzz won't be available if raqm isn't available.
        expected_features -= {"raqm", "fribidi", "harfbuzz"}

    assert set(features.get_supported_features()) == expected_features



from __future__ import annotations

import sys
from pathlib import Path

import pytest

from PIL import Image

# This test is not run automatically.
#
# It requires > 2gb memory for the >2 gigapixel image generated in the
# second test.  Running this automatically would amount to a denial of
# service on our testing infrastructure.  I expect this test to fail
# on any 32-bit machine, as well as any smallish things (like
# Raspberry Pis).


np = pytest.importorskip("numpy", reason="NumPy not installed")

YDIM = 32769
XDIM = 48000


pytestmark = pytest.mark.skipif(sys.maxsize <= 2**32, reason="requires 64-bit system")


def _write_png(tmp_path: Path, xdim: int, ydim: int) -> None:
    dtype = np.uint8
    a = np.zeros((xdim, ydim), dtype=dtype)
    f = tmp_path / "temp.png"
    im = Image.fromarray(a, "L")
    im.save(f)


def test_large(tmp_path: Path) -> None:
    """succeeded prepatch"""
    _write_png(tmp_path, XDIM, YDIM)


def test_2gpx(tmp_path: Path) -> None:
    """failed prepatch"""
    _write_png(tmp_path, XDIM, XDIM)



from __future__ import annotations

from PIL import Image, ImageDraw, ImageFont

from .helper import skip_unless_feature


class TestFontCrash:
    def _fuzz_font(self, font: ImageFont.FreeTypeFont) -> None:
        # Copy of the code from fuzz_font() in Tests/oss-fuzz/fuzzers.py
        # that triggered a problem when fuzzing
        font.getbbox("ABC")
        font.getmask("test text")
        with Image.new(mode="RGBA", size=(200, 200)) as im:
            draw = ImageDraw.Draw(im)
            draw.multiline_textbbox((10, 10), "ABC\nAaaa", font, stroke_width=2)
            draw.text((10, 10), "Test Text", font=font, fill="#000")

    @skip_unless_feature("freetype2")
    def test_segfault(self) -> None:
        font = ImageFont.truetype("Tests/fonts/fuzz_font-5203009437302784")
        self._fuzz_font(font)



from __future__ import annotations

from io import BytesIO
from pathlib import Path

import pytest

from PIL import Image, SgiImagePlugin

from .helper import (
    assert_image_equal,
    assert_image_equal_tofile,
    assert_image_similar,
    hopper,
)


def test_rgb() -> None:
    # Created with ImageMagick then renamed:
    # convert hopper.ppm -compress None sgi:hopper.rgb
    test_file = "Tests/images/hopper.rgb"

    with Image.open(test_file) as im:
        assert_image_equal(im, hopper())
        assert im.get_format_mimetype() == "image/rgb"


def test_rgb16() -> None:
    assert_image_equal_tofile(hopper(), "Tests/images/hopper16.rgb")


def test_l() -> None:
    # Created with ImageMagick
    # convert hopper.ppm -monochrome -compress None sgi:hopper.bw
    test_file = "Tests/images/hopper.bw"

    with Image.open(test_file) as im:
        assert_image_similar(im, hopper("L"), 2)
        assert im.get_format_mimetype() == "image/sgi"


def test_rgba() -> None:
    # Created with ImageMagick:
    # convert transparent.png -compress None transparent.sgi
    test_file = "Tests/images/transparent.sgi"

    with Image.open(test_file) as im:
        assert_image_equal_tofile(im, "Tests/images/transparent.png")
        assert im.get_format_mimetype() == "image/sgi"


def test_rle() -> None:
    # Created with ImageMagick:
    # convert hopper.ppm  hopper.sgi
    test_file = "Tests/images/hopper.sgi"

    with Image.open(test_file) as im:
        assert_image_equal_tofile(im, "Tests/images/hopper.rgb")


def test_rle16() -> None:
    test_file = "Tests/images/tv16.sgi"

    with Image.open(test_file) as im:
        assert_image_equal_tofile(im, "Tests/images/tv.rgb")


def test_invalid_file() -> None:
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(ValueError):
        SgiImagePlugin.SgiImageFile(invalid_file)


def test_unsupported_image_mode() -> None:
    with open("Tests/images/hopper.rgb", "rb") as fp:
        data = fp.read()
    data = data[:3] + b"\x03" + data[4:]
    with pytest.raises(ValueError, match="Unsupported SGI image mode"):
        with Image.open(BytesIO(data)):
            pass


def roundtrip(img: Image.Image, tmp_path: Path) -> None:
    out = tmp_path / "temp.sgi"
    img.save(out, format="sgi")
    assert_image_equal_tofile(img, out)

    out = tmp_path / "fp.sgi"
    with open(out, "wb") as fp:
        img.save(fp)
        assert_image_equal_tofile(img, out)

        assert not fp.closed


@pytest.mark.parametrize("mode", ("L", "RGB", "RGBA"))
def test_write(mode: str, tmp_path: Path) -> None:
    roundtrip(hopper(mode), tmp_path)


def test_write_L_mode_1_dimension(tmp_path: Path) -> None:
    roundtrip(Image.new("L", (10, 1)), tmp_path)


def test_write16(tmp_path: Path) -> None:
    test_file = "Tests/images/hopper16.rgb"

    with Image.open(test_file) as im:
        out = tmp_path / "temp.sgi"
        im.save(out, format="sgi", bpc=2)

        assert_image_equal_tofile(im, out)


def test_unsupported_mode(tmp_path: Path) -> None:
    im = hopper("LA")
    out = tmp_path / "temp.sgi"

    with pytest.raises(ValueError):
        im.save(out, format="sgi")


def test_unsupported_number_of_bytes_per_pixel(tmp_path: Path) -> None:
    im = hopper()
    out = tmp_path / "temp.sgi"

    with pytest.raises(ValueError, match="Unsupported number of bytes per pixel"):
        im.save(out, bpc=3)



from __future__ import annotations

import locale

import pytest

from PIL import Image

# ref https://github.com/python-pillow/Pillow/issues/272
# on windows, in polish locale:

# import locale
# print(locale.setlocale(locale.LC_ALL, 'polish'))
# import string
# print(len(string.whitespace))
# print(ord(string.whitespace[6]))

# Polish_Poland.1250
# 7
# 160

# one of string.whitespace is not freely convertible into ascii.

path = "Tests/images/hopper.jpg"


def test_sanity() -> None:
    with Image.open(path):
        pass
    try:
        locale.setlocale(locale.LC_ALL, "polish")
    except locale.Error:
        pytest.skip("Polish locale not available")

    try:
        with Image.open(path):
            pass
    finally:
        locale.setlocale(locale.LC_ALL, (None, None))



from __future__ import annotations

import pytest

from PIL import Image

from .helper import (
    assert_image_equal,
    assert_image_similar,
    fromstring,
    hopper,
    skip_unless_feature,
    tostring,
)


def test_sanity() -> None:
    im = hopper()
    im.thumbnail((100, 100))

    assert im.size == (100, 100)


def test_aspect() -> None:
    im = Image.new("L", (128, 128))
    im.thumbnail((100, 100))
    assert im.size == (100, 100)

    im = Image.new("L", (128, 256))
    im.thumbnail((100, 100))
    assert im.size == (50, 100)

    im = Image.new("L", (128, 256))
    im.thumbnail((50, 100))
    assert im.size == (50, 100)

    im = Image.new("L", (256, 128))
    im.thumbnail((100, 100))
    assert im.size == (100, 50)

    im = Image.new("L", (256, 128))
    im.thumbnail((100, 50))
    assert im.size == (100, 50)

    im = Image.new("L", (64, 64))
    im.thumbnail((100, 100))
    assert im.size == (64, 64)

    im = Image.new("L", (256, 162))  # ratio is 1.5802469136
    im.thumbnail((33, 33))
    assert im.size == (33, 21)  # ratio is 1.5714285714

    im = Image.new("L", (162, 256))  # ratio is 0.6328125
    im.thumbnail((33, 33))
    assert im.size == (21, 33)  # ratio is 0.6363636364

    im = Image.new("L", (145, 100))  # ratio is 1.45
    im.thumbnail((50, 50))
    assert im.size == (50, 34)  # ratio is 1.47058823529

    im = Image.new("L", (100, 145))  # ratio is 0.689655172414
    im.thumbnail((50, 50))
    assert im.size == (34, 50)  # ratio is 0.68

    im = Image.new("L", (100, 30))  # ratio is 3.333333333333
    im.thumbnail((75, 75))
    assert im.size == (75, 23)  # ratio is 3.260869565217


def test_division_by_zero() -> None:
    im = Image.new("L", (200, 2))
    im.thumbnail((75, 75))
    assert im.size == (75, 1)


def test_float() -> None:
    im = Image.new("L", (128, 128))
    im.thumbnail((99.9, 99.9))
    assert im.size == (99, 99)


def test_no_resize() -> None:
    # Check that draft() can resize the image to the destination size
    with Image.open("Tests/images/hopper.jpg") as im:
        im.draft(None, (64, 64))
        assert im.size == (64, 64)

    # Test thumbnail(), where only draft() is necessary to resize the image
    with Image.open("Tests/images/hopper.jpg") as im:
        im.thumbnail((64, 64))
        assert im.size == (64, 64)


@skip_unless_feature("libtiff")
def test_transposed() -> None:
    with Image.open("Tests/images/g4_orientation_5.tif") as im:
        assert im.size == (590, 88)

        im.thumbnail((64, 64))
        assert im.size == (64, 10)

    with Image.open("Tests/images/g4_orientation_5.tif") as im:
        im.thumbnail((590, 88), reducing_gap=None)
        assert im.size == (590, 88)


def test_load_first_unless_jpeg(monkeypatch: pytest.MonkeyPatch) -> None:
    # Test that thumbnail() still uses draft() for JPEG
    with Image.open("Tests/images/hopper.jpg") as im:
        original_draft = im.draft

        def im_draft(
            mode: str | None, size: tuple[int, int] | None
        ) -> tuple[str, tuple[int, int, float, float]] | None:
            result = original_draft(mode, size)
            assert result is not None

            return result

        monkeypatch.setattr(im, "draft", im_draft)

        im.thumbnail((64, 64))


# valgrind test is failing with memory allocated in libjpeg
@pytest.mark.valgrind_known_error(reason="Known Failing")
def test_DCT_scaling_edges() -> None:
    # Make an image with red borders and size (N * 8) + 1 to cross DCT grid
    im = Image.new("RGB", (257, 257), "red")
    im.paste(Image.new("RGB", (235, 235)), (11, 11))

    thumb = fromstring(tostring(im, "JPEG", quality=99, subsampling=0))
    # small reducing_gap to amplify the effect
    thumb.thumbnail((32, 32), Image.Resampling.BICUBIC, reducing_gap=1.0)

    ref = im.resize((32, 32), Image.Resampling.BICUBIC)
    # This is still JPEG, some error is present. Without the fix it is 11.5
    assert_image_similar(thumb, ref, 1.5)


def test_reducing_gap_values() -> None:
    im = hopper()
    im.thumbnail((18, 18), Image.Resampling.BICUBIC)

    ref = hopper()
    ref.thumbnail((18, 18), Image.Resampling.BICUBIC, reducing_gap=2.0)
    # reducing_gap=2.0 should be the default
    assert_image_equal(ref, im)

    ref = hopper()
    ref.thumbnail((18, 18), Image.Resampling.BICUBIC, reducing_gap=None)
    with pytest.raises(pytest.fail.Exception):
        assert_image_equal(ref, im)

    assert_image_similar(ref, im, 3.5)


def test_reducing_gap_for_DCT_scaling() -> None:
    with Image.open("Tests/images/hopper.jpg") as ref:
        # thumbnail should call draft with reducing_gap scale
        ref.draft(None, (18 * 3, 18 * 3))
        im_ref = ref.resize((18, 18), Image.Resampling.BICUBIC)

        with Image.open("Tests/images/hopper.jpg") as im:
            im.thumbnail((18, 18), Image.Resampling.BICUBIC, reducing_gap=3.0)

            assert_image_similar(im_ref, im, 1.4)



from __future__ import annotations

import math
from fractions import Fraction
from pathlib import Path

import pytest

from PIL import Image, TiffImagePlugin
from PIL.TiffImagePlugin import IFDRational

from .helper import hopper, skip_unless_feature


def _test_equal(
    num: float | Fraction | IFDRational,
    denom: int,
    target: float | Fraction | IFDRational,
) -> None:
    t = IFDRational(num, denom)

    assert target == t
    assert t == target


def test_sanity() -> None:
    _test_equal(1, 1, 1)
    _test_equal(1, 1, Fraction(1, 1))

    _test_equal(2, 2, 1)
    _test_equal(1.0, 1, Fraction(1, 1))

    _test_equal(Fraction(1, 1), 1, Fraction(1, 1))
    _test_equal(IFDRational(1, 1), 1, 1)

    _test_equal(1, 2, Fraction(1, 2))
    _test_equal(1, 2, IFDRational(1, 2))

    _test_equal(7, 5, 1.4)


def test_float() -> None:
    # float() must agree with ==, int() and repr(), which all use the
    # normalized fraction
    r = IFDRational(1.5, 3)
    assert r == 0.5
    assert float(r) == 0.5

    # Integer numerator and 0/0 (nan) cases
    assert float(IFDRational(4, 2)) == 2.0
    assert math.isnan(float(IFDRational(0, 0)))


def test_ranges() -> None:
    for num in range(1, 10):
        for denom in range(1, 10):
            assert IFDRational(num, denom) == IFDRational(num, denom)


def test_nonetype() -> None:
    # Fails if the _delegate function doesn't return a valid function

    xres = IFDRational(72)
    yres = IFDRational(72)
    assert xres._val is not None
    assert xres.numerator is not None
    assert xres.denominator is not None
    assert yres._val is not None

    assert xres
    assert yres


@pytest.mark.parametrize(
    "libtiff", (pytest.param(True, marks=skip_unless_feature("libtiff")), False)
)
def test_ifd_rational_save(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path, libtiff: bool
) -> None:
    im = hopper()
    out = tmp_path / "temp.tiff"
    res = IFDRational(301, 1)

    monkeypatch.setattr(TiffImagePlugin, "WRITE_LIBTIFF", libtiff)
    im.save(out, dpi=(res, res), compression="raw")

    with Image.open(out) as reloaded:
        assert isinstance(reloaded, TiffImagePlugin.TiffImageFile)
        assert float(IFDRational(301, 1)) == float(reloaded.tag_v2[282])



from __future__ import annotations

from io import BytesIO

from PIL import WalImageFile

from .helper import assert_image_equal_tofile

TEST_FILE = "Tests/images/hopper.wal"


def test_open() -> None:
    with WalImageFile.open(TEST_FILE) as im:
        assert im.format == "WAL"
        assert im.format_description == "Quake2 Texture"
        assert im.mode == "P"
        assert im.size == (128, 128)
        assert "next_name" not in im.info

        assert isinstance(im, WalImageFile.WalImageFile)

        assert_image_equal_tofile(im, "Tests/images/hopper_wal.png")


def test_next_name() -> None:
    with open(TEST_FILE, "rb") as fp:
        data = bytearray(fp.read())
    data[56:60] = b"Test"
    f = BytesIO(data)
    with WalImageFile.open(f) as im:
        assert im.info["next_name"] == b"Test"


def test_load() -> None:
    with WalImageFile.open(TEST_FILE) as im:
        px = im.load()
        assert px is not None
        assert px[0, 0] == 122

        # Test again now that it has already been loaded once
        px = im.load()
        assert px is not None
        assert px[0, 0] == 122



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from types import ModuleType

import pytest

from PIL import Image, WebPImagePlugin

from .helper import mark_if_feature_version, skip_unless_feature

pytestmark = skip_unless_feature("webp")

ElementTree: ModuleType | None
try:
    from defusedxml import ElementTree
except ImportError:
    ElementTree = None


def test_read_exif_metadata() -> None:
    file_path = "Tests/images/flower.webp"
    with Image.open(file_path) as image:
        assert isinstance(image, WebPImagePlugin.WebPImageFile)
        assert image.format == "WEBP"
        exif_data = image.info.get("exif", None)
        assert exif_data

        exif = image._getexif()
        assert exif is not None

        # Camera make
        assert exif[271] == "Canon"

        with Image.open("Tests/images/flower.jpg") as jpeg_image:
            expected_exif = jpeg_image.info["exif"]

            assert exif_data == expected_exif


def test_read_exif_metadata_without_prefix() -> None:
    with Image.open("Tests/images/flower2.webp") as im:
        # Assert prefix is not present
        assert not im.info["exif"].startswith(b"Exif\x00\x00")

        exif = im.getexif()
        assert exif[305] == "Adobe Photoshop CS6 (Macintosh)"


@mark_if_feature_version(
    pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
)
def test_write_exif_metadata() -> None:
    file_path = "Tests/images/flower.jpg"
    test_buffer = BytesIO()
    with Image.open(file_path) as image:
        expected_exif = image.info["exif"]

        image.save(test_buffer, "webp", exif=expected_exif)

    test_buffer.seek(0)
    with Image.open(test_buffer) as webp_image:
        webp_exif = webp_image.info.get("exif", None)
    assert webp_exif == expected_exif[6:], "WebP EXIF didn't match"


def test_read_icc_profile() -> None:
    file_path = "Tests/images/flower2.webp"
    with Image.open(file_path) as image:
        assert image.format == "WEBP"
        assert image.info.get("icc_profile", None)

        icc = image.info["icc_profile"]

        with Image.open("Tests/images/flower2.jpg") as jpeg_image:
            expected_icc = jpeg_image.info["icc_profile"]

            assert icc == expected_icc


@mark_if_feature_version(
    pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
)
def test_write_icc_metadata() -> None:
    file_path = "Tests/images/flower2.jpg"
    test_buffer = BytesIO()
    with Image.open(file_path) as image:
        expected_icc_profile = image.info["icc_profile"]

        image.save(test_buffer, "webp", icc_profile=expected_icc_profile)

    test_buffer.seek(0)
    with Image.open(test_buffer) as webp_image:
        webp_icc_profile = webp_image.info.get("icc_profile", None)

    assert webp_icc_profile
    if webp_icc_profile:
        assert webp_icc_profile == expected_icc_profile, "Webp ICC didn't match"


@mark_if_feature_version(
    pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
)
def test_read_no_exif() -> None:
    file_path = "Tests/images/flower.jpg"
    test_buffer = BytesIO()
    with Image.open(file_path) as image:
        assert "exif" in image.info

        image.save(test_buffer, "webp")

    test_buffer.seek(0)
    with Image.open(test_buffer) as webp_image:
        assert isinstance(webp_image, WebPImagePlugin.WebPImageFile)
        assert not webp_image._getexif()


def test_getxmp() -> None:
    with Image.open("Tests/images/flower.webp") as im:
        assert "xmp" not in im.info
        if ElementTree is None:
            with pytest.warns(
                UserWarning,
                match="XMP data cannot be read without defusedxml dependency",
            ):
                xmp = im.getxmp()
        else:
            xmp = im.getxmp()
        assert xmp == {}

    with Image.open("Tests/images/flower2.webp") as im:
        if ElementTree is None:
            with pytest.warns(
                UserWarning,
                match="XMP data cannot be read without defusedxml dependency",
            ):
                assert im.getxmp() == {}
        else:
            assert "xmp" in im.info
            assert (
                im.getxmp()["xmpmeta"]["xmptk"]
                == "Adobe XMP Core 5.3-c011 66.145661, 2012/02/06-14:56:27        "
            )


def test_write_animated_metadata(tmp_path: Path) -> None:
    iccp_data = b"<iccp_data>"
    exif_data = b"<exif_data>"
    xmp_data = b"<xmp_data>"

    temp_file = tmp_path / "temp.webp"
    with Image.open("Tests/images/anim_frame1.webp") as frame1:
        with Image.open("Tests/images/anim_frame2.webp") as frame2:
            frame1.save(
                temp_file,
                save_all=True,
                append_images=[frame2, frame1, frame2],
                icc_profile=iccp_data,
                exif=exif_data,
                xmp=xmp_data,
            )

    with Image.open(temp_file) as image:
        assert "icc_profile" in image.info
        assert "exif" in image.info
        assert "xmp" in image.info
        assert iccp_data == image.info.get("icc_profile", None)
        assert exif_data == image.info.get("exif", None)
        assert xmp_data == image.info.get("xmp", None)



from __future__ import annotations

from pathlib import Path

import pytest

from PIL import Image, ImageSequence, PsdImagePlugin, TiffImagePlugin

from .helper import assert_image_equal, hopper, skip_unless_feature


def test_sanity(tmp_path: Path) -> None:
    test_file = tmp_path / "temp.im"

    im = hopper("RGB")
    im.save(test_file)

    seq = ImageSequence.Iterator(im)

    index = 0
    for frame in seq:
        assert_image_equal(im, frame)
        assert im.tell() == index
        index += 1

    assert index == 1

    with pytest.raises(AttributeError):
        ImageSequence.Iterator(0)  # type: ignore[arg-type]


def test_iterator() -> None:
    with Image.open("Tests/images/multipage.tiff") as im:
        assert isinstance(im, TiffImagePlugin.TiffImageFile)
        i = ImageSequence.Iterator(im)
        for index in range(im.n_frames):
            assert i[index] == next(i)
        with pytest.raises(IndexError):
            i[index + 1]
        with pytest.raises(StopIteration):
            next(i)


def test_iterator_min_frame() -> None:
    with Image.open("Tests/images/hopper.psd") as im:
        assert isinstance(im, PsdImagePlugin.PsdImageFile)
        i = ImageSequence.Iterator(im)
        for index in range(1, im.n_frames):
            assert i[index] == next(i)


@pytest.mark.parametrize(
    "libtiff", (pytest.param(True, marks=skip_unless_feature("libtiff")), False)
)
def test_multipage_tiff(monkeypatch: pytest.MonkeyPatch, libtiff: bool) -> None:
    monkeypatch.setattr(TiffImagePlugin, "READ_LIBTIFF", libtiff)
    with Image.open("Tests/images/multipage.tiff") as im:
        for index, frame in enumerate(ImageSequence.Iterator(im)):
            frame.load()
            assert index == im.tell()
            frame.convert("RGB")


def test_consecutive() -> None:
    with Image.open("Tests/images/multipage.tiff") as im:
        first_frame = None
        for frame in ImageSequence.Iterator(im):
            if first_frame is None:
                first_frame = frame.copy()
        assert first_frame is not None
        for frame in ImageSequence.Iterator(im):
            assert_image_equal(frame, first_frame)
            break


def test_palette_mmap() -> None:
    # Using mmap in ImageFile can require to reload the palette.
    with Image.open("Tests/images/multipage-mmap.tiff") as im:
        palette = im.getpalette()
        assert palette is not None
        color1 = palette[:3]
        im.seek(0)

        palette = im.getpalette()
        assert palette is not None
        color2 = palette[:3]
        assert color1 == color2


def test_all_frames() -> None:
    # Test a single image
    with Image.open("Tests/images/iss634.gif") as im:
        ims = ImageSequence.all_frames(im)

        assert len(ims) == 42
        for i, im_frame in enumerate(ims):
            assert im_frame is not im

            im.seek(i)
            assert_image_equal(im, im_frame)

        # Test a series of images
        ims = ImageSequence.all_frames([im, hopper(), im])
        assert len(ims) == 85

        # Test an operation
        ims = ImageSequence.all_frames(im, lambda im_frame: im_frame.rotate(90))
        for i, im_frame in enumerate(ims):
            im.seek(i)
            assert_image_equal(im.rotate(90), im_frame)



from __future__ import annotations

import os
import re
import warnings
from io import BytesIO
from pathlib import Path
from types import ModuleType
from typing import Any, cast

import pytest

from PIL import (
    ExifTags,
    Image,
    ImageFile,
    ImageOps,
    JpegImagePlugin,
    UnidentifiedImageError,
    features,
)

from .helper import (
    assert_image,
    assert_image_equal,
    assert_image_equal_tofile,
    assert_image_similar,
    assert_image_similar_tofile,
    djpeg_available,
    hopper,
    is_win32,
    mark_if_feature_version,
    skip_unless_feature,
    timeout_unless_slower_valgrind,
)

ElementTree: ModuleType | None
try:
    from defusedxml import ElementTree
except ImportError:
    ElementTree = None

TEST_FILE = "Tests/images/hopper.jpg"


@skip_unless_feature("jpg")
class TestFileJpeg:
    def roundtrip_with_bytes(
        self, im: Image.Image, **options: Any
    ) -> tuple[JpegImagePlugin.JpegImageFile, int]:
        out = BytesIO()
        im.save(out, "JPEG", **options)
        test_bytes = out.tell()
        out.seek(0)
        reloaded = cast(JpegImagePlugin.JpegImageFile, Image.open(out))
        return reloaded, test_bytes

    def roundtrip(
        self, im: Image.Image, **options: Any
    ) -> JpegImagePlugin.JpegImageFile:
        return self.roundtrip_with_bytes(im, **options)[0]

    def gen_random_image(self, size: tuple[int, int], mode: str = "RGB") -> Image.Image:
        """Generates a very hard to compress file
        :param size: tuple
        :param mode: optional image mode

        """
        return Image.frombytes(mode, size, os.urandom(size[0] * size[1] * len(mode)))

    def test_sanity(self) -> None:
        # internal version number
        version = features.version_codec("jpg")
        assert version is not None
        assert re.search(r"\d+\.\d+$", version)

        with Image.open(TEST_FILE) as im:
            im.load()
            assert im.mode == "RGB"
            assert im.size == (128, 128)
            assert im.format == "JPEG"
            assert im.get_format_mimetype() == "image/jpeg"

    @pytest.mark.parametrize("size", ((1, 0), (0, 1), (0, 0)))
    def test_zero(self, size: tuple[int, int], tmp_path: Path) -> None:
        f = tmp_path / "temp.jpg"
        im = Image.new("RGB", size)
        with pytest.raises(ValueError, match="cannot write empty image"):
            im.save(f)

    def test_app(self) -> None:
        # Test APP/COM reader (@PIL135)
        with Image.open(TEST_FILE) as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            assert im.applist[0] == ("APP0", b"JFIF\x00\x01\x01\x01\x00`\x00`\x00\x00")
            assert im.applist[1] == (
                "COM",
                b"File written by Adobe Photoshop\xa8 4.0\x00",
            )
            assert len(im.applist) == 2

            assert im.info["comment"] == b"File written by Adobe Photoshop\xa8 4.0\x00"
            assert im.app["COM"] == im.info["comment"]

    def test_comment_write(self) -> None:
        with Image.open(TEST_FILE) as im:
            assert im.info["comment"] == b"File written by Adobe Photoshop\xa8 4.0\x00"

            # Test that existing comment is saved by default
            out = BytesIO()
            im.save(out, format="JPEG")
            with Image.open(out) as reloaded:
                assert im.info["comment"] == reloaded.info["comment"]

            # Ensure that a blank comment causes any existing comment to be removed
            for comment in ("", b"", None):
                out = BytesIO()
                im.save(out, format="JPEG", comment=comment)
                with Image.open(out) as reloaded:
                    assert "comment" not in reloaded.info

            # Test that a comment argument overrides the default comment
            for comment in ("Test comment text", b"Test comment text"):
                out = BytesIO()
                im.save(out, format="JPEG", comment=comment)
                with Image.open(out) as reloaded:
                    assert reloaded.info["comment"] == b"Test comment text"

    def test_cmyk(self) -> None:
        # Test CMYK handling.  Thanks to Tim and Charlie for test data,
        # Michael for getting me to look one more time.
        def check(im: ImageFile.ImageFile) -> None:
            cmyk = im.getpixel((0, 0))
            assert isinstance(cmyk, tuple)
            c, m, y, k = (x / 255.0 for x in cmyk)
            assert c == 0.0
            assert m > 0.8
            assert y > 0.8
            assert k == 0.0
            # the opposite corner is black
            cmyk = im.getpixel((im.size[0] - 1, im.size[1] - 1))
            assert isinstance(cmyk, tuple)
            k = cmyk[3] / 255.0
            assert k > 0.9

        with Image.open("Tests/images/pil_sample_cmyk.jpg") as im:
            # the source image has red pixels in the upper left corner.
            check(im)

            # roundtrip, and check again
            check(self.roundtrip(im))

    def test_rgb(self) -> None:
        def getchannels(im: JpegImagePlugin.JpegImageFile) -> tuple[int, ...]:
            return tuple(v[0] for v in im.layer)

        im = hopper()
        im_ycbcr = self.roundtrip(im)
        assert getchannels(im_ycbcr) == (1, 2, 3)
        assert_image_similar(im, im_ycbcr, 17)

        im_rgb = self.roundtrip(im, keep_rgb=True)
        assert getchannels(im_rgb) == (ord("R"), ord("G"), ord("B"))
        assert_image_similar(im, im_rgb, 12)

    @pytest.mark.parametrize(
        "test_image_path",
        [TEST_FILE, "Tests/images/pil_sample_cmyk.jpg"],
    )
    def test_dpi(self, test_image_path: str) -> None:
        def test(xdpi: int, ydpi: int | None = None) -> tuple[int, int] | None:
            with Image.open(test_image_path) as im:
                im = self.roundtrip(im, dpi=(xdpi, ydpi or xdpi))
            return im.info.get("dpi")

        assert test(72) == (72, 72)
        assert test(300) == (300, 300)
        assert test(100, 200) == (100, 200)
        assert test(0) is None  # square pixels

    def test_dpi_jfif_cm(self) -> None:
        with Image.open("Tests/images/jfif_unit_cm.jpg") as im:
            assert im.info["dpi"] == (2.54, 5.08)

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_icc(self, tmp_path: Path) -> None:
        # Test ICC support
        with Image.open("Tests/images/rgb.jpg") as im1:
            icc_profile = im1.info["icc_profile"]
            assert len(icc_profile) == 3144
            # Roundtrip via physical file.
            f = tmp_path / "temp.jpg"
            im1.save(f, icc_profile=icc_profile)
        with Image.open(f) as im2:
            assert im2.info.get("icc_profile") == icc_profile
            # Roundtrip via memory buffer.
            im1 = self.roundtrip(hopper())
            im2 = self.roundtrip(hopper(), icc_profile=icc_profile)
            assert_image_equal(im1, im2)
            assert not im1.info.get("icc_profile")
            assert im2.info.get("icc_profile")

    @pytest.mark.parametrize(
        "n",
        (
            0,
            1,
            3,
            4,
            5,
            65533 - 14,  # full JPEG marker block
            65533 - 14 + 1,  # full block plus one byte
            ImageFile.MAXBLOCK,  # full buffer block
            ImageFile.MAXBLOCK + 1,  # full buffer block plus one byte
            ImageFile.MAXBLOCK * 4 + 3,  # large block
        ),
    )
    def test_icc_big(self, n: int) -> None:
        # Make sure that the "extra" support handles large blocks
        # The ICC APP marker can store 65519 bytes per marker, so
        # using a 4-byte test code should allow us to detect out of
        # order issues.
        icc_profile = (b"Test" * int(n / 4 + 1))[:n]
        assert len(icc_profile) == n  # sanity
        im1 = self.roundtrip(hopper(), icc_profile=icc_profile)
        assert im1.info.get("icc_profile") == (icc_profile or None)

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_large_icc_meta(self, tmp_path: Path) -> None:
        # https://github.com/python-pillow/Pillow/issues/148
        # Sometimes the meta data on the icc_profile block is bigger than
        # Image.MAXBLOCK or the image size.
        with Image.open("Tests/images/icc_profile_big.jpg") as im:
            f = tmp_path / "temp.jpg"
            icc_profile = im.info["icc_profile"]
            # Should not raise OSError for image with icc larger than image size.
            im.save(
                f,
                progressive=True,
                quality=95,
                icc_profile=icc_profile,
                optimize=True,
            )

        with Image.open("Tests/images/flower2.jpg") as im:
            f = tmp_path / "temp2.jpg"
            im.save(f, progressive=True, quality=94, icc_profile=b" " * 53955)

        with Image.open("Tests/images/flower2.jpg") as im:
            f = tmp_path / "temp3.jpg"
            im.save(f, progressive=True, quality=94, exif=b" " * 43668)

    def test_optimize(self) -> None:
        im1, im1_bytes = self.roundtrip_with_bytes(hopper())
        im2, im2_bytes = self.roundtrip_with_bytes(hopper(), optimize=0)
        im3, im3_bytes = self.roundtrip_with_bytes(hopper(), optimize=1)
        assert_image_equal(im1, im2)
        assert_image_equal(im1, im3)
        assert im1_bytes >= im2_bytes
        assert im1_bytes >= im3_bytes

    def test_optimize_large_buffer(self, tmp_path: Path) -> None:
        # https://github.com/python-pillow/Pillow/issues/148
        f = tmp_path / "temp.jpg"
        # this requires ~ 1.5x Image.MAXBLOCK
        im = Image.new("RGB", (4096, 4096), 0xFF3333)
        im.save(f, format="JPEG", optimize=True)

    def test_progressive(self) -> None:
        im1, im1_bytes = self.roundtrip_with_bytes(hopper())
        im2 = self.roundtrip(hopper(), progressive=False)
        im3, im3_bytes = self.roundtrip_with_bytes(hopper(), progressive=True)
        assert not im1.info.get("progressive")
        assert not im2.info.get("progressive")
        assert im3.info.get("progressive")

        if features.check_feature("mozjpeg"):
            assert_image_similar(im1, im3, 9.39)
        else:
            assert_image_equal(im1, im3)
        assert im1_bytes >= im3_bytes

    def test_progressive_large_buffer(self, tmp_path: Path) -> None:
        f = tmp_path / "temp.jpg"
        # this requires ~ 1.5x Image.MAXBLOCK
        im = Image.new("RGB", (4096, 4096), 0xFF3333)
        im.save(f, format="JPEG", progressive=True)

    def test_progressive_large_buffer_highest_quality(self, tmp_path: Path) -> None:
        f = tmp_path / "temp.jpg"
        im = self.gen_random_image((255, 255))
        # this requires more bytes than pixels in the image
        im.save(f, format="JPEG", progressive=True, quality=100)

    def test_progressive_cmyk_buffer(self) -> None:
        # Issue 2272, quality 90 cmyk image is tripping the large buffer bug.
        f = BytesIO()
        im = self.gen_random_image((256, 256), "CMYK")
        im.save(f, format="JPEG", progressive=True, quality=94)

    def test_large_exif(self, tmp_path: Path) -> None:
        # https://github.com/python-pillow/Pillow/issues/148
        f = tmp_path / "temp.jpg"
        im = hopper()
        im.save(f, "JPEG", quality=90, exif=b"1" * 65533)

        with pytest.raises(ValueError):
            im.save(f, "JPEG", quality=90, exif=b"1" * 65534)

    def test_exif_typeerror(self) -> None:
        with Image.open("Tests/images/exif_typeerror.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)

            # Should not raise a TypeError
            im._getexif()

    def test_exif_gps(self, tmp_path: Path) -> None:
        expected_exif_gps = {
            0: b"\x00\x00\x00\x01",
            2: 4294967295,
            5: b"\x01",
            30: 65535,
            29: "1999:99:99 99:99:99",
        }
        gps_index = 34853

        # Reading
        with Image.open("Tests/images/exif_gps.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            exif_data = im._getexif()
            assert exif_data is not None
            assert exif_data[gps_index] == expected_exif_gps

        # Writing
        f = tmp_path / "temp.jpg"
        exif = Image.Exif()
        exif[gps_index] = expected_exif_gps
        hopper().save(f, exif=exif)

        with Image.open(f) as reloaded:
            assert isinstance(reloaded, JpegImagePlugin.JpegImageFile)
            exif_data = reloaded._getexif()
            assert exif_data is not None
            assert exif_data[gps_index] == expected_exif_gps

    def test_empty_exif_gps(self) -> None:
        with Image.open("Tests/images/empty_gps_ifd.jpg") as im:
            exif = im.getexif()
            del exif[0x8769]

            # Assert that it needs to be transposed
            assert exif[0x0112] == Image.Transpose.TRANSVERSE

            # Assert that the GPS IFD is present and empty
            assert exif.get_ifd(0x8825) == {}

            transposed = ImageOps.exif_transpose(im)
        exif = transposed.getexif()
        assert exif.get_ifd(0x8825) == {}

        # Assert that it was transposed
        assert 0x0112 not in exif

    def test_exif_equality(self) -> None:
        # In 7.2.0, Exif rationals were changed to be read as
        # TiffImagePlugin.IFDRational. This class had a bug in __eq__,
        # breaking the self-equality of Exif data
        exifs = []
        for i in range(2):
            with Image.open("Tests/images/exif-200dpcm.jpg") as im:
                assert isinstance(im, JpegImagePlugin.JpegImageFile)
                exifs.append(im._getexif())
        assert exifs[0] == exifs[1]

    def test_exif_rollback(self) -> None:
        # rolling back exif support in 3.1 to pre-3.0 formatting.
        # expected from 2.9, with b/u qualifiers switched for 3.2 compatibility
        # this test passes on 2.9 and 3.1, but not 3.0
        expected_exif = {
            34867: 4294967295,
            258: (24, 24, 24),
            36867: "2099:09:29 10:10:10",
            34853: {
                0: b"\x00\x00\x00\x01",
                2: 4294967295,
                5: b"\x01",
                30: 65535,
                29: "1999:99:99 99:99:99",
            },
            296: 65535,
            34665: 185,
            41994: 65535,
            514: 4294967295,
            271: "Make",
            272: "XXX-XXX",
            305: "PIL",
            42034: (1, 1, 1, 1),
            42035: "LensMake",
            34856: b"\xaa\xaa\xaa\xaa\xaa\xaa",
            282: 4294967295,
            33434: 4294967295,
        }

        with Image.open("Tests/images/exif_gps.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            exif = im._getexif()
        assert exif is not None

        for tag, value in expected_exif.items():
            assert value == exif[tag]

    def test_exif_gps_typeerror(self) -> None:
        with Image.open("Tests/images/exif_gps_typeerror.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)

            # Should not raise a TypeError
            im._getexif()

    def test_progressive_compat(self) -> None:
        im1 = self.roundtrip(hopper())
        assert not im1.info.get("progressive")
        assert not im1.info.get("progression")

        im2 = self.roundtrip(hopper(), progressive=0)
        im3 = self.roundtrip(hopper(), progression=0)  # compatibility
        assert not im2.info.get("progressive")
        assert not im2.info.get("progression")
        assert not im3.info.get("progressive")
        assert not im3.info.get("progression")

        im2 = self.roundtrip(hopper(), progressive=1)
        im3 = self.roundtrip(hopper(), progression=1)  # compatibility
        if features.check_feature("mozjpeg"):
            assert_image_similar(im1, im2, 9.39)
            assert_image_similar(im1, im3, 9.39)
        else:
            assert_image_equal(im1, im2)
            assert_image_equal(im1, im3)
        assert im2.info.get("progressive")
        assert im2.info.get("progression")
        assert im3.info.get("progressive")
        assert im3.info.get("progression")

    def test_quality(self) -> None:
        im1, im1_bytes = self.roundtrip_with_bytes(hopper())
        im2, im2_bytes = self.roundtrip_with_bytes(hopper(), quality=50)
        assert_image(im1, im2.mode, im2.size)
        assert im1_bytes >= im2_bytes

        im3, im3_bytes = self.roundtrip_with_bytes(hopper(), quality=0)
        assert_image(im1, im3.mode, im3.size)
        assert im2_bytes > im3_bytes

    def test_smooth(self) -> None:
        im1 = self.roundtrip(hopper())
        im2 = self.roundtrip(hopper(), smooth=100)
        assert_image(im1, im2.mode, im2.size)

    def test_subsampling(self) -> None:
        def getsampling(
            im: JpegImagePlugin.JpegImageFile,
        ) -> tuple[int, int, int, int, int, int]:
            layer = im.layer
            return layer[0][1:3] + layer[1][1:3] + layer[2][1:3]

        # experimental API
        for subsampling in (-1, 3):  # (default, invalid)
            im = self.roundtrip(hopper(), subsampling=subsampling)
            assert getsampling(im) == (2, 2, 1, 1, 1, 1)
        for subsampling1 in (0, "4:4:4"):
            im = self.roundtrip(hopper(), subsampling=subsampling1)
            assert getsampling(im) == (1, 1, 1, 1, 1, 1)
        for subsampling1 in (1, "4:2:2"):
            im = self.roundtrip(hopper(), subsampling=subsampling1)
            assert getsampling(im) == (2, 1, 1, 1, 1, 1)
        for subsampling1 in (2, "4:2:0", "4:1:1"):
            im = self.roundtrip(hopper(), subsampling=subsampling1)
            assert getsampling(im) == (2, 2, 1, 1, 1, 1)

        # RGB colorspace
        for subsampling1 in (-1, 0, "4:4:4"):
            # "4:4:4" doesn't really make sense for RGB, but the conversion
            # to an integer happens at a higher level
            im = self.roundtrip(hopper(), keep_rgb=True, subsampling=subsampling1)
            assert getsampling(im) == (1, 1, 1, 1, 1, 1)
        for subsampling1 in (1, "4:2:2", 2, "4:2:0", 3):
            with pytest.raises(OSError):
                self.roundtrip(hopper(), keep_rgb=True, subsampling=subsampling1)

        with pytest.raises(TypeError):
            self.roundtrip(hopper(), subsampling="1:1:1")

    def test_exif(self) -> None:
        with Image.open("Tests/images/pil_sample_rgb.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            info = im._getexif()
            assert info is not None
            assert info[305] == "Adobe Photoshop CS Macintosh"

    def test_get_child_images(self) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            ims = im.get_child_images()

        assert len(ims) == 1
        assert_image_similar_tofile(ims[0], "Tests/images/flower_thumbnail.png", 2.1)

    def test_mp(self) -> None:
        with Image.open("Tests/images/pil_sample_rgb.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            assert im._getmp() is None

    def test_quality_keep(self, tmp_path: Path) -> None:
        # RGB
        with Image.open("Tests/images/hopper.jpg") as im:
            f = tmp_path / "temp.jpg"
            im.save(f, quality="keep")
        # Grayscale
        with Image.open("Tests/images/hopper_gray.jpg") as im:
            f = tmp_path / "temp.jpg"
            im.save(f, quality="keep")
        # CMYK
        with Image.open("Tests/images/pil_sample_cmyk.jpg") as im:
            f = tmp_path / "temp.jpg"
            im.save(f, quality="keep")

    def test_junk_jpeg_header(self) -> None:
        # https://github.com/python-pillow/Pillow/issues/630
        filename = "Tests/images/junk_jpeg_header.jpg"
        with Image.open(filename):
            pass

    def test_ff00_jpeg_header(self) -> None:
        filename = "Tests/images/jpeg_ff00_header.jpg"
        with Image.open(filename):
            pass

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_truncated_jpeg_should_read_all_the_data(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        filename = "Tests/images/truncated_jpeg.jpg"
        monkeypatch.setattr(ImageFile, "LOAD_TRUNCATED_IMAGES", True)
        with Image.open(filename) as im:
            im.load()
            assert im.getbbox() is not None

    def test_truncated_jpeg_throws_oserror(self) -> None:
        filename = "Tests/images/truncated_jpeg.jpg"
        with Image.open(filename) as im:
            with pytest.raises(OSError):
                im.load()

            # Test that the error is raised if loaded a second time
            with pytest.raises(OSError):
                im.load()

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_qtables(self) -> None:
        def _n_qtables_helper(n: int, test_file: str) -> None:
            b = BytesIO()
            with Image.open(test_file) as im:
                im.save(b, "JPEG", qtables=[[n] * 64] * n)
            with Image.open(b) as im:
                assert isinstance(im, JpegImagePlugin.JpegImageFile)
                assert len(im.quantization) == n
                reloaded = self.roundtrip(im, qtables="keep")
                assert im.quantization == reloaded.quantization
                assert max(reloaded.quantization[0]) <= 255

        with Image.open("Tests/images/hopper.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            qtables = im.quantization
            reloaded = self.roundtrip(im, qtables=qtables, subsampling=0)
            assert im.quantization == reloaded.quantization
            assert_image_similar(im, self.roundtrip(im, qtables="web_low"), 30)
            assert_image_similar(im, self.roundtrip(im, qtables="web_high"), 30)
            assert_image_similar(im, self.roundtrip(im, qtables="keep"), 30)

            # valid bounds for baseline qtable
            bounds_qtable = [int(s) for s in ("255 1 " * 32).split(None)]
            im2 = self.roundtrip(im, qtables=[bounds_qtable])
            assert im2.quantization == {0: bounds_qtable}

            # values from wizard.txt in jpeg9-a src package.
            standard_l_qtable = [int(s) for s in """
                16  11  10  16  24  40  51  61
                12  12  14  19  26  58  60  55
                14  13  16  24  40  57  69  56
                14  17  22  29  51  87  80  62
                18  22  37  56  68 109 103  77
                24  35  55  64  81 104 113  92
                49  64  78  87 103 121 120 101
                72  92  95  98 112 100 103  99
                """.split(None)]

            standard_chrominance_qtable = [int(s) for s in """
                17  18  24  47  99  99  99  99
                18  21  26  66  99  99  99  99
                24  26  56  99  99  99  99  99
                47  66  99  99  99  99  99  99
                99  99  99  99  99  99  99  99
                99  99  99  99  99  99  99  99
                99  99  99  99  99  99  99  99
                99  99  99  99  99  99  99  99
                """.split(None)]

            for quality in range(101):
                qtable_from_qtable_quality = self.roundtrip(
                    im,
                    qtables={0: standard_l_qtable, 1: standard_chrominance_qtable},
                    quality=quality,
                ).quantization

                qtable_from_quality = self.roundtrip(im, quality=quality).quantization

                if features.check_feature("libjpeg_turbo"):
                    assert qtable_from_qtable_quality == qtable_from_quality
                else:
                    assert qtable_from_qtable_quality[0] == qtable_from_quality[0]
                    assert (
                        qtable_from_qtable_quality[1][1:] == qtable_from_quality[1][1:]
                    )

            # list of qtable lists
            assert_image_similar(
                im,
                self.roundtrip(
                    im, qtables=[standard_l_qtable, standard_chrominance_qtable]
                ),
                30,
            )

            # tuple of qtable lists
            assert_image_similar(
                im,
                self.roundtrip(
                    im, qtables=(standard_l_qtable, standard_chrominance_qtable)
                ),
                30,
            )

            # dict of qtable lists
            assert_image_similar(
                im,
                self.roundtrip(
                    im, qtables={0: standard_l_qtable, 1: standard_chrominance_qtable}
                ),
                30,
            )

            _n_qtables_helper(1, "Tests/images/hopper_gray.jpg")
            _n_qtables_helper(1, "Tests/images/pil_sample_rgb.jpg")
            _n_qtables_helper(2, "Tests/images/pil_sample_rgb.jpg")
            _n_qtables_helper(3, "Tests/images/pil_sample_rgb.jpg")
            _n_qtables_helper(1, "Tests/images/pil_sample_cmyk.jpg")
            _n_qtables_helper(2, "Tests/images/pil_sample_cmyk.jpg")
            _n_qtables_helper(3, "Tests/images/pil_sample_cmyk.jpg")
            _n_qtables_helper(4, "Tests/images/pil_sample_cmyk.jpg")

            # not a sequence
            with pytest.raises(ValueError):
                self.roundtrip(im, qtables="a")
            # sequence wrong length
            with pytest.raises(ValueError):
                self.roundtrip(im, qtables=[])
            # sequence wrong length
            with pytest.raises(ValueError):
                self.roundtrip(im, qtables=[1, 2, 3, 4, 5])

            # qtable entry not a sequence
            with pytest.raises(ValueError):
                self.roundtrip(im, qtables=[1])
            # qtable entry has wrong number of items
            with pytest.raises(ValueError):
                self.roundtrip(im, qtables=[[1, 2, 3, 4]])

    def test_load_16bit_qtables(self) -> None:
        with Image.open("Tests/images/hopper_16bit_qtables.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            assert len(im.quantization) == 2
            assert len(im.quantization[0]) == 64
            assert max(im.quantization[0]) > 255

    def test_save_multiple_16bit_qtables(self) -> None:
        with Image.open("Tests/images/hopper_16bit_qtables.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            im2 = self.roundtrip(im, qtables="keep")
            assert im.quantization == im2.quantization

    def test_save_single_16bit_qtable(self) -> None:
        with Image.open("Tests/images/hopper_16bit_qtables.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            im2 = self.roundtrip(im, qtables={0: im.quantization[0]})
            assert len(im2.quantization) == 1
            assert im2.quantization[0] == im.quantization[0]

    def test_save_low_quality_baseline_qtables(self) -> None:
        with Image.open(TEST_FILE) as im:
            im2 = self.roundtrip(im, quality=10)
            assert len(im2.quantization) == 2
            assert max(im2.quantization[0]) <= 255
            assert max(im2.quantization[1]) <= 255

    @pytest.mark.parametrize(
        "blocks, rows, markers",
        ((0, 0, 0), (1, 0, 15), (3, 0, 5), (8, 0, 1), (0, 1, 3), (0, 2, 1)),
    )
    def test_restart_markers(self, blocks: int, rows: int, markers: int) -> None:
        im = Image.new("RGB", (32, 32))  # 16 MCUs
        out = BytesIO()
        im.save(
            out,
            format="JPEG",
            restart_marker_blocks=blocks,
            restart_marker_rows=rows,
            # force 8x8 pixel MCUs
            subsampling=0,
        )
        assert len(re.findall(b"\xff[\xd0-\xd7]", out.getvalue())) == markers

    @pytest.mark.skipif(not djpeg_available(), reason="djpeg not available")
    def test_load_djpeg(self) -> None:
        with Image.open(TEST_FILE) as img:
            assert isinstance(img, JpegImagePlugin.JpegImageFile)
            img.load_djpeg()
            assert_image_similar_tofile(img, TEST_FILE, 5)

    def test_no_duplicate_0x1001_tag(self) -> None:
        # Arrange
        tag_ids = {v: k for k, v in ExifTags.TAGS.items()}

        # Assert
        assert tag_ids["RelatedImageWidth"] == 0x1001
        assert tag_ids["RelatedImageLength"] == 0x1002

    def test_MAXBLOCK_scaling(self, tmp_path: Path) -> None:
        im = self.gen_random_image((512, 512))
        f = tmp_path / "temp.jpeg"
        im.save(f, quality=100, optimize=True)

        with Image.open(f) as reloaded:
            # none of these should crash
            reloaded.save(f, quality="keep")
            reloaded.save(f, quality="keep", progressive=True)
            reloaded.save(f, quality="keep", optimize=True)

    def test_bad_mpo_header(self) -> None:
        """Treat unknown MPO as JPEG"""
        # Arrange

        # Act
        # Shouldn't raise error
        with pytest.warns(UserWarning, match="malformed MPO file"):
            im = Image.open("Tests/images/sugarshack_bad_mpo_header.jpg")

        # Assert
        assert im.format == "JPEG"

        im.close()

    @pytest.mark.parametrize("mode", ("1", "L", "RGB", "RGBX", "CMYK", "YCbCr"))
    def test_save_correct_modes(self, mode: str) -> None:
        out = BytesIO()
        img = Image.new(mode, (20, 20))
        img.save(out, "JPEG")

    @pytest.mark.parametrize("mode", ("LA", "La", "RGBA", "RGBa", "P"))
    def test_save_wrong_modes(self, mode: str) -> None:
        # ref https://github.com/python-pillow/Pillow/issues/2005
        out = BytesIO()
        img = Image.new(mode, (20, 20))
        with pytest.raises(OSError):
            img.save(out, "JPEG")

    def test_save_tiff_with_dpi(self, tmp_path: Path) -> None:
        # Arrange
        outfile = tmp_path / "temp.tif"
        with Image.open("Tests/images/hopper.tif") as im:
            # Act
            im.save(outfile, "JPEG", dpi=im.info["dpi"])

            # Assert
            with Image.open(outfile) as reloaded:
                reloaded.load()
                assert im.info["dpi"] == reloaded.info["dpi"]

    def test_save_dpi_rounding(self, tmp_path: Path) -> None:
        outfile = tmp_path / "temp.jpg"
        with Image.open("Tests/images/hopper.jpg") as im:
            im.save(outfile, dpi=(72.2, 72.2))

            with Image.open(outfile) as reloaded:
                assert reloaded.info["dpi"] == (72, 72)

            im.save(outfile, dpi=(72.8, 72.8))

        with Image.open(outfile) as reloaded:
            assert reloaded.info["dpi"] == (73, 73)

    def test_dpi_tuple_from_exif(self) -> None:
        # Arrange
        # This Photoshop CC 2017 image has DPI in EXIF not metadata
        # EXIF XResolution is (2000000, 10000)
        with Image.open("Tests/images/photoshop-200dpi.jpg") as im:
            # Act / Assert
            assert im.info.get("dpi") == (200, 200)

    def test_dpi_int_from_exif(self) -> None:
        # Arrange
        # This image has DPI in EXIF not metadata
        # EXIF XResolution is 72
        with Image.open("Tests/images/exif-72dpi-int.jpg") as im:
            # Act / Assert
            assert im.info.get("dpi") == (72, 72)

    def test_dpi_from_dpcm_exif(self) -> None:
        # Arrange
        # This is photoshop-200dpi.jpg with EXIF resolution unit set to cm:
        # exiftool -exif:ResolutionUnit=cm photoshop-200dpi.jpg
        with Image.open("Tests/images/exif-200dpcm.jpg") as im:
            # Act / Assert
            assert im.info.get("dpi") == (508, 508)

    def test_dpi_exif_zero_division(self) -> None:
        # Arrange
        # This is photoshop-200dpi.jpg with EXIF resolution set to 0/0:
        # exiftool -XResolution=0/0 -YResolution=0/0 photoshop-200dpi.jpg
        with Image.open("Tests/images/exif-dpi-zerodivision.jpg") as im:
            # Act / Assert
            # This should return the default, and not raise a ZeroDivisionError
            assert im.info.get("dpi") == (72, 72)

    def test_dpi_exif_string(self) -> None:
        # Arrange
        # 0x011A tag in this exif contains string '300300\x02'
        with Image.open("Tests/images/broken_exif_dpi.jpg") as im:
            # Act / Assert
            # This should return the default
            assert im.info.get("dpi") == (72, 72)

    def test_dpi_exif_truncated(self) -> None:
        # Arrange
        with Image.open("Tests/images/truncated_exif_dpi.jpg") as im:
            # Act / Assert
            # This should return the default
            assert im.info.get("dpi") == (72, 72)

    def test_no_dpi_in_exif(self) -> None:
        # Arrange
        # This is photoshop-200dpi.jpg with resolution removed from EXIF:
        # exiftool "-*resolution*"= photoshop-200dpi.jpg
        with Image.open("Tests/images/no-dpi-in-exif.jpg") as im:
            # Act / Assert
            # "When the image resolution is unknown, 72 [dpi] is designated."
            # https://exiv2.org/tags.html
            assert im.info.get("dpi") == (72, 72)

    def test_invalid_exif(self) -> None:
        # This is no-dpi-in-exif with the tiff header of the exif block
        # hexedited from MM * to FF FF FF FF
        with Image.open("Tests/images/invalid-exif.jpg") as im:
            # This should return the default, and not a SyntaxError or
            # OSError for unidentified image.
            assert im.info.get("dpi") == (72, 72)

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_exif_x_resolution(self, tmp_path: Path) -> None:
        with Image.open("Tests/images/flower.jpg") as im:
            exif = im.getexif()
            assert exif[282] == 180

            out = tmp_path / "out.jpg"
            with warnings.catch_warnings():
                warnings.simplefilter("error")

                im.save(out, exif=exif)

        with Image.open(out) as reloaded:
            assert reloaded.getexif()[282] == 180

    def test_invalid_exif_x_resolution(self) -> None:
        # When no x or y resolution is defined in EXIF
        with Image.open("Tests/images/invalid-exif-without-x-resolution.jpg") as im:
            # This should return the default, and not a ValueError or
            # OSError for an unidentified image.
            assert im.info.get("dpi") == (72, 72)

    def test_ifd_offset_exif(self) -> None:
        # Arrange
        # This image has been manually hexedited to have an IFD offset of 10,
        # in contrast to normal 8
        with Image.open("Tests/images/exif-ifd-offset.jpg") as im:
            # Act / Assert
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            exif = im._getexif()
            assert exif is not None
            assert exif[306] == "2017:03:13 23:03:09"

    def test_multiple_exif(self) -> None:
        with Image.open("Tests/images/multiple_exif.jpg") as im:
            assert im.getexif()[270] == "firstsecond"

    @mark_if_feature_version(
        pytest.mark.valgrind_known_error, "libjpeg_turbo", "2.0", reason="Known Failing"
    )
    def test_photoshop(self) -> None:
        with Image.open("Tests/images/photoshop-200dpi.jpg") as im:
            assert im.info["photoshop"][0x03ED] == {
                "XResolution": 200.0,
                "DisplayedUnitsX": 1,
                "YResolution": 200.0,
                "DisplayedUnitsY": 1,
            }

            # Test that the image can still load, even with broken Photoshop data
            # This image had the APP13 length hexedited to be smaller
            assert_image_equal_tofile(im, "Tests/images/photoshop-200dpi-broken.jpg")

        # This image does not contain a Photoshop header string
        with Image.open("Tests/images/app13.jpg") as im:
            assert "photoshop" not in im.info

    def test_photoshop_malformed_and_multiple(self) -> None:
        with Image.open("Tests/images/app13-multiple.jpg") as im:
            assert isinstance(im, JpegImagePlugin.JpegImageFile)
            assert "photoshop" in im.info
            assert 24 == len(im.info["photoshop"])
            apps_13_lengths = [len(v) for k, v in im.applist if k == "APP13"]
            assert [65504, 24] == apps_13_lengths

    def test_adobe_transform(self) -> None:
        with Image.open("Tests/images/pil_sample_rgb.jpg") as im:
            assert im.info["adobe_transform"] == 1

        with Image.open("Tests/images/pil_sample_cmyk.jpg") as im:
            assert im.info["adobe_transform"] == 2

        # This image has been manually hexedited
        # so that the APP14 reports its length to be 11,
        # leaving no room for "adobe_transform"
        with Image.open("Tests/images/truncated_app14.jpg") as im:
            assert "adobe" in im.info
            assert "adobe_transform" not in im.info

    def test_icc_after_SOF(self) -> None:
        with Image.open("Tests/images/icc-after-SOF.jpg") as im:
            assert im.info["icc_profile"] == b"profile"

    def test_jpeg_magic_number(self, monkeypatch: pytest.MonkeyPatch) -> None:
        size = 4097
        buffer = BytesIO(b"\xff" * size)  # Many xff bytes
        max_pos = 0
        orig_read = buffer.read

        def read(n: int | None = -1) -> bytes:
            nonlocal max_pos
            res = orig_read(n)
            max_pos = max(max_pos, buffer.tell())
            return res

        monkeypatch.setattr(buffer, "read", read)
        with pytest.raises(UnidentifiedImageError):
            with Image.open(buffer):
                pass

        # Assert the entire file has not been read
        assert 0 < max_pos < size

    def test_getxmp(self) -> None:
        with Image.open("Tests/images/xmp_test.jpg") as im:
            if ElementTree is None:
                with pytest.warns(
                    UserWarning,
                    match="XMP data cannot be read without defusedxml dependency",
                ):
                    assert im.getxmp() == {}
            else:
                assert "xmp" in im.info
                xmp = im.getxmp()

                description = xmp["xmpmeta"]["RDF"]["Description"]
                assert description["DerivedFrom"] == {
                    "documentID": "8367D410E636EA95B7DE7EBA1C43A412",
                    "originalDocumentID": "8367D410E636EA95B7DE7EBA1C43A412",
                }
                assert description["Look"]["Description"]["Group"]["Alt"]["li"] == {
                    "lang": "x-default",
                    "text": "Profiles",
                }
                assert description["ToneCurve"]["Seq"]["li"] == ["0, 0", "255, 255"]

                # Attribute
                assert description["Version"] == "10.4"

        if ElementTree is not None:
            with Image.open("Tests/images/hopper.jpg") as im:
                assert im.getxmp() == {}

    def test_getxmp_no_prefix(self) -> None:
        with Image.open("Tests/images/xmp_no_prefix.jpg") as im:
            if ElementTree is None:
                with pytest.warns(
                    UserWarning,
                    match="XMP data cannot be read without defusedxml dependency",
                ):
                    assert im.getxmp() == {}
            else:
                assert im.getxmp() == {"xmpmeta": {"key": "value"}}

    def test_getxmp_padded(self) -> None:
        with Image.open("Tests/images/xmp_padded.jpg") as im:
            if ElementTree is None:
                with pytest.warns(
                    UserWarning,
                    match="XMP data cannot be read without defusedxml dependency",
                ):
                    assert im.getxmp() == {}
            else:
                assert im.getxmp() == {"xmpmeta": None}

    def test_save_xmp(self, tmp_path: Path) -> None:
        f = tmp_path / "temp.jpg"
        im = hopper()
        im.save(f, xmp=b"XMP test")
        with Image.open(f) as reloaded:
            assert reloaded.info["xmp"] == b"XMP test"

            # Check that XMP is not saved from image info
            reloaded.save(f)

        with Image.open(f) as reloaded:
            assert "xmp" not in reloaded.info

        im.save(f, xmp=b"1" * 65504)
        with Image.open(f) as reloaded:
            assert reloaded.info["xmp"] == b"1" * 65504

        with pytest.raises(ValueError):
            im.save(f, xmp=b"1" * 65505)

    @timeout_unless_slower_valgrind(1)
    def test_eof(self, monkeypatch: pytest.MonkeyPatch) -> None:
        # Even though this decoder never says that it is finished
        # the image should still end when there is no new data
        class InfiniteMockPyDecoder(ImageFile.PyDecoder):
            def decode(self, buffer: Image.DecoderInput) -> tuple[int, int]:
                return 0, 0

        Image.register_decoder("INFINITE", InfiniteMockPyDecoder)

        with Image.open(TEST_FILE) as im:
            im.tile = [
                ImageFile._Tile("INFINITE", (0, 0, 128, 128), 0, ("RGB", 0, 1)),
            ]
            monkeypatch.setattr(ImageFile, "LOAD_TRUNCATED_IMAGES", True)
            im.load()

    def test_separate_tables(self) -> None:
        im = hopper()
        data = []  # [interchange, tables-only, image-only]
        for streamtype in range(3):
            out = BytesIO()
            im.save(out, format="JPEG", streamtype=streamtype)
            data.append(out.getvalue())

        # SOI, EOI
        for marker in b"\xff\xd8", b"\xff\xd9":
            assert marker in data[1]
            assert marker in data[2]

        # DQT
        markers = [b"\xff\xdb"]
        if features.check_feature("libjpeg_turbo"):
            # DHT
            markers.append(b"\xff\xc4")
        for marker in markers:
            assert marker in data[1]
            assert marker not in data[2]

        # SOF0, SOS, APP0 (JFIF header)
        for marker in b"\xff\xc0", b"\xff\xda", b"\xff\xe0":
            assert marker not in data[1]
            assert marker in data[2]

        with Image.open(BytesIO(data[0])) as interchange_im:
            with Image.open(BytesIO(data[1] + data[2])) as combined_im:
                assert_image_equal(interchange_im, combined_im)

    def test_repr_jpeg(self) -> None:
        im = hopper()
        b = im._repr_jpeg_()
        assert b is not None

        with Image.open(BytesIO(b)) as repr_jpeg:
            assert repr_jpeg.format == "JPEG"
            assert_image_similar(im, repr_jpeg, 17)

    def test_repr_jpeg_error_returns_none(self) -> None:
        im = hopper("F")

        assert im._repr_jpeg_() is None


@pytest.mark.skipif(not is_win32(), reason="Windows only")
@skip_unless_feature("jpg")
class TestFileCloseW32:
    def test_fd_leak(self, tmp_path: Path) -> None:
        tmpfile = tmp_path / "temp.jpg"

        with Image.open("Tests/images/hopper.jpg") as im:
            im.save(tmpfile)

        im = Image.open(tmpfile)
        assert im.fp is not None
        assert not im.fp.closed
        fp = im.fp
        with pytest.raises(OSError):
            os.remove(tmpfile)
        im.load()
        assert fp.closed
        # this should not fail, as load should have closed the file.
        os.remove(tmpfile)



from __future__ import annotations

from typing import Any

import pytest

from PIL import Image, ImageMath, _imagingmath


def pixel(im: Image.Image | int) -> str | int:
    if isinstance(im, int):
        return int(im)  # hack to deal with booleans

    return f"{im.mode} {repr(im.getpixel((0, 0)))}"


A = Image.new("L", (1, 1), 1)
B = Image.new("L", (1, 1), 2)
Z = Image.new("L", (1, 1), 0)  # Z for zero
F = Image.new("F", (1, 1), 3)
I = Image.new("I", (1, 1), 4)  # noqa: E741

A2 = A.resize((2, 2))
B2 = B.resize((2, 2))

images: dict[str, Any] = {"A": A, "B": B, "F": F, "I": I}


def test_sanity() -> None:
    assert ImageMath.lambda_eval(lambda args: 1) == 1
    assert ImageMath.lambda_eval(lambda args: 1 + args["A"], A=2) == 3
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] + args["B"], A=A, B=B))
        == "I 3"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] + args["B"], **images))
        == "I 3"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["float"](args["A"]) + args["B"], **images
            )
        )
        == "F 3.0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["int"](args["float"](args["A"]) + args["B"]), **images
            )
        )
        == "I 3"
    )


def test_ops() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] * -1, **images)) == "I -1"

    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] + args["B"], **images))
        == "I 3"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] - args["B"], **images))
        == "I -1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] * args["B"], **images))
        == "I 2"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] / args["B"], **images))
        == "I 0"
    )
    assert pixel(ImageMath.lambda_eval(lambda args: args["B"] ** 2, **images)) == "I 4"
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] ** 33, **images))
        == "I 2147483647"
    )

    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["float"](args["A"]) + args["B"], **images
            )
        )
        == "F 3.0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["float"](args["A"]) - args["B"], **images
            )
        )
        == "F -1.0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["float"](args["A"]) * args["B"], **images
            )
        )
        == "F 2.0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["float"](args["A"]) / args["B"], **images
            )
        )
        == "F 0.5"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(lambda args: args["float"](args["B"]) ** 2, **images)
        )
        == "F 4.0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(lambda args: args["float"](args["B"]) ** 33, **images)
        )
        == "F 8589934592.0"
    )


def test_logical() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: not args["A"], **images)) == 0
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] and args["B"], **images))
        == "L 2"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] or args["B"], **images))
        == "L 1"
    )


def test_convert() -> None:
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["convert"](args["A"] + args["B"], "L"), **images
            )
        )
        == "L 3"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["convert"](args["A"] + args["B"], "1"), **images
            )
        )
        == "1 0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["convert"](args["A"] + args["B"], "RGB"), **images
            )
        )
        == "RGB (3, 3, 3)"
    )


def test_compare() -> None:
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["min"](args["A"], args["B"]), **images
            )
        )
        == "I 1"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["max"](args["A"], args["B"]), **images
            )
        )
        == "I 2"
    )
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] == 1, **images)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] == 2, **images)) == "I 0"


def test_one_image_larger() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] + args["B"], A=A2, B=B))
        == "I 3"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] + args["B"], A=A, B=B2))
        == "I 3"
    )


def test_abs() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: abs(args["A"]), A=A)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: abs(args["B"]), B=B)) == "I 2"


def test_binary_mod() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] % args["A"], A=A)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] % args["B"], B=B)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] % args["B"], A=A, B=B))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] % args["A"], A=A, B=B))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] % args["A"], A=A, Z=Z))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] % args["B"], B=B, Z=Z))
        == "I 0"
    )


def test_bitwise_invert() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: ~args["Z"], Z=Z)) == "I -1"
    assert pixel(ImageMath.lambda_eval(lambda args: ~args["A"], A=A)) == "I -2"
    assert pixel(ImageMath.lambda_eval(lambda args: ~args["B"], B=B)) == "I -3"


def test_bitwise_and() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] & args["Z"], A=A, Z=Z))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] & args["A"], A=A, Z=Z))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] & args["Z"], A=A, Z=Z))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] & args["A"], A=A, Z=Z))
        == "I 1"
    )


def test_bitwise_or() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] | args["Z"], A=A, Z=Z))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] | args["A"], A=A, Z=Z))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] | args["Z"], A=A, Z=Z))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] | args["A"], A=A, Z=Z))
        == "I 1"
    )


def test_bitwise_xor() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] ^ args["Z"], A=A, Z=Z))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["Z"] ^ args["A"], A=A, Z=Z))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] ^ args["Z"], A=A, Z=Z))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] ^ args["A"], A=A, Z=Z))
        == "I 0"
    )


def test_bitwise_leftshift() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: args["Z"] << 0, Z=Z)) == "I 0"
    assert pixel(ImageMath.lambda_eval(lambda args: args["Z"] << 1, Z=Z)) == "I 0"
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] << 0, A=A)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] << 1, A=A)) == "I 2"


def test_bitwise_rightshift() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: args["Z"] >> 0, Z=Z)) == "I 0"
    assert pixel(ImageMath.lambda_eval(lambda args: args["Z"] >> 1, Z=Z)) == "I 0"
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] >> 0, A=A)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: args["A"] >> 1, A=A)) == "I 0"


def test_logical_eq() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] == args["A"], A=A)) == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] == args["B"], B=B)) == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] == args["B"], A=A, B=B))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] == args["A"], A=A, B=B))
        == "I 0"
    )


def test_logical_ne() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] != args["A"], A=A)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] != args["B"], B=B)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] != args["B"], A=A, B=B))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] != args["A"], A=A, B=B))
        == "I 1"
    )


def test_logical_lt() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] < args["A"], A=A)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] < args["B"], B=B)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] < args["B"], A=A, B=B))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] < args["A"], A=A, B=B))
        == "I 0"
    )


def test_logical_le() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] <= args["A"], A=A)) == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] <= args["B"], B=B)) == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] <= args["B"], A=A, B=B))
        == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] <= args["A"], A=A, B=B))
        == "I 0"
    )


def test_logical_gt() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] > args["A"], A=A)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] > args["B"], B=B)) == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] > args["B"], A=A, B=B))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] > args["A"], A=A, B=B))
        == "I 1"
    )


def test_logical_ge() -> None:
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] >= args["A"], A=A)) == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] >= args["B"], B=B)) == "I 1"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["A"] >= args["B"], A=A, B=B))
        == "I 0"
    )
    assert (
        pixel(ImageMath.lambda_eval(lambda args: args["B"] >= args["A"], A=A, B=B))
        == "I 1"
    )


def test_logical_equal() -> None:
    assert (
        pixel(
            ImageMath.lambda_eval(lambda args: args["equal"](args["A"], args["A"]), A=A)
        )
        == "I 1"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(lambda args: args["equal"](args["B"], args["B"]), B=B)
        )
        == "I 1"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(lambda args: args["equal"](args["Z"], args["Z"]), Z=Z)
        )
        == "I 1"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["equal"](args["A"], args["B"]), A=A, B=B
            )
        )
        == "I 0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["equal"](args["B"], args["A"]), A=A, B=B
            )
        )
        == "I 0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["equal"](args["A"], args["Z"]), A=A, Z=Z
            )
        )
        == "I 0"
    )


def test_logical_not_equal() -> None:
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["notequal"](args["A"], args["A"]), A=A
            )
        )
        == "I 0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["notequal"](args["B"], args["B"]), B=B
            )
        )
        == "I 0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["notequal"](args["Z"], args["Z"]), Z=Z
            )
        )
        == "I 0"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["notequal"](args["A"], args["B"]), A=A, B=B
            )
        )
        == "I 1"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["notequal"](args["B"], args["A"]), A=A, B=B
            )
        )
        == "I 1"
    )
    assert (
        pixel(
            ImageMath.lambda_eval(
                lambda args: args["notequal"](args["A"], args["Z"]), A=A, Z=Z
            )
        )
        == "I 1"
    )


def test_reflected_operands() -> None:
    assert pixel(ImageMath.lambda_eval(lambda args: 1 + args["A"], **images)) == "I 2"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 - args["A"], **images)) == "I 0"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 * args["A"], **images)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 / args["A"], **images)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 % args["A"], **images)) == "I 0"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 ** args["A"], **images)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 & args["A"], **images)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 | args["A"], **images)) == "I 1"
    assert pixel(ImageMath.lambda_eval(lambda args: 1 ^ args["A"], **images)) == "I 0"


def test_unsupported_mode() -> None:
    im = Image.new("RGB", (1, 1))
    with pytest.raises(ValueError, match="unsupported mode: RGB"):
        ImageMath.lambda_eval(lambda args: args["im"] + 1, im=im)


def test_bad_operand_type(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delattr(_imagingmath, "abs_I")
    with pytest.raises(TypeError, match="bad operand type for 'abs'"):
        ImageMath.lambda_eval(lambda args: abs(args["I"]), I=I)

    monkeypatch.delattr(_imagingmath, "max_F")
    with pytest.raises(TypeError, match="bad operand type for 'max'"):
        ImageMath.lambda_eval(lambda args: args["max"](args["I"], args["F"]), I=I, F=F)
